;emuterm.asm
;
;include <srdefn>
;
; Simple ASCII/APL Terminal Emulator
;
; Special Features:
; 1. XON/XOFF support for transmission to host (HostSync)
; 2. XON/XOFF support for reception from host (TTSync)
; 3. Break generation using STOP key
; 4. Scroll control using Shift/Keypad 0 key
; 5. Emulation exit using Shift/Keypad 8 key
; 6. Character set (ASCII/APL) support
;    using Shift In (SI) / Shift Out (SO) characters
; 7. CTRL key support using RVS key and other key (@,A,B,...,_)
;    (keys must be pressed one after the other, not simultaneously)
;
; Known Problems:
; 1. Wrap at column 80 may require host setup
; 2. Holding down the SHIFT key generates continual interrupts
;    such that incoming characters may be lost.
; 3. The duration of the BREAK interval in SRBreak_ may need
;    to be increased for certain systems.
;
; To link this emulator, create and use the following command file:
;
;       term
;       org $A00
;       include "disk/1.watlib.exp"
;       export ACIAHOST_ = $EFF0
;       "emuterm.b09"
;       "srget.b09"
;       "srinth.b09"
;       "srirq.b09"
;       "srput.b09"
;       "srstuff.b09"
;       "srtie.b09"
;
        xref    Openf_
        xref    FgetChar_
        xref    Closef_
        xref    PutChar_
        xref    TBreak_
        xref    TSetChar_

        xref    SRTie_
        xref    SRConnect_
        xref    SRDisconnect_
        xref    SRGetChar_
        xref    SRPutChar_
        xref    SRBreak_

Echo    equ     $7F     ; Use location $7F for Echo control
Service equ     $32     ; System Service code
Buffer_ equ     $400    ; Buffer for storing received characters
Screen_ equ     $8000   ; Location of screen memory
SKD_    equ     $10E    ; Location of screen / keyboard descriptor
                        ; offsets from SKD_
CURSOR  equ     20      ; cursor : int
APLSET  equ     26      ; aplset : bool

KBDEL   equ     $04     ; Keyboard DEL
KBESC   equ     $06     ; Keyboard ESC
CTRL_G  equ     $07     ; ASCII Bell
CURUP   equ     $0B     ; SuperPET Cursor Up
CR      equ     $0D     ; Carriaage Return
CTRL_N  equ     $0E     ; ASCII Shift Out
CTRL_O  equ     $0F     ; ASCII Shift In
CTRL_Q  equ     $11     ; ASCII DC1 (XON)
CTRL_S  equ     $13     ; ASCII DC3 (XOFF)
ESCAPE  equ     $1B     ; ASCII Escape
RUBOUT  equ     $7F     ; ASCII DEL (Rubout)
KBCTRL  equ     $FF     ; Keyboard OFF/RVS

PFKEY   equ     $80             ; Function keypad
KBEXIT  equ     PFKEY + 8       ; Keypad 8 (shifted)
KBHOLD  equ     PFKEY + 10      ; Keypad 0 (shifted)

ASCII   equ     1               ; ASCII character set selection
APL     equ     2               ; APL character set selection

xdef    EmuTerm_
EmuTerm_  equ  *
                                ; * Start of parameter section
                                ;   A more sophisticated terminal emulator
                                ;   would allow User setting of the following:
        LDB     #$00            ;   FALSE value (patchable)
        STB     Echo            ;   Disable local echoing of typed characters
        LDB     #$FF            ;   TRUE value (patchable)
        STB     HostSync        ;   Recognize host XON/XOFF synchronization
        LDB     #$FF            ;   TRUE value (patchable)
        STB     TTSync          ;   Perform terminal XON/XOFF synchronization
                                ; * end of parameter section
        CLRB                    ; FALSE value
        STB     Hold            ; Scroll control is off
        STB     PrevKey         ; Last key typed is initialized NULL
                                ; Openf ( "KEYBOARD", "R" )
        LDD     #ReadMode       ; pointer to open mode string
        PSHS    D               ; push parameter 2
        LDD     #KeyBoard       ; pointer to file name
        JSR     Openf_          ; open keyboard
        LEAS    2,S             ; pop parameter
        STD     KBIn            ; save keyboard file pointer
        LDD     #Buffer_        ; get address of buffer
        JSR     SRTie_          ; SRTie( Buffer)
        JSR     SRConnect_      ; SRConnect()
        LDB     #CTRL_Q         ; SRPutChar( CTRL_Q )
        JSR     SRPutChar_      ; transmit XON in case host is in XOFF state
        loop                    ; loop (talk to host computer)
          JSR   TBreak_         ; - check for STOP key
          TSTB                  ; - if TBreak()
          if ne                 ; - then
            JSR SRBreak_        ; - - SRBreak()
          endif                 ; - endif
          LDD   KBIn            ; - Key := FGetChar (KBIn)
          JSR   FGetChar_       ; -
          STB   Key             ; -
          CMPB  #KBEXIT         ; - quit if key = KBEXIT
          quif  eq              ; -
          guess                 ; - guess
            CMPB #KBCTRL        ; - - quit if Key <> KBCTRL
            quif ne             ; - -
            STB  PrevKey        ; - - remember CTRL toggle
          admit                 ; - admit
            CMPB #KBHOLD        ; - - quit if Key <> KBHOLD
            quif ne             ; - -
            LDA  #$FF           ; - -
            STA  TTSync         ; - - TTSync := TRUE
            COM  Hold           ; - - Hold := ~ Hold
          admit                 ; - admit
            TSTB                ; - - quit if Key = NULLCHAR
            quif eq             ; - -
            CLRA                ; - -
            STA  Hold           ; - - Hold := FALSE
            BSR  TranOut        ; - - translate key before sending to host
            TST  Echo           ; - - if local echoing
            if   ne             ; - - then
              BSR  LocalEcho    ; - - - LocalEcho( Key )
            endif               ; - - endif
            STB  PrevKey        ; - - PrecKey := Key
            JSR  SRPutChar_     ; - - SRPutChar( Key )
          endguess              ; - endguess
          TST   Hold            ; - if ~ Hold
          if    eq              ; - then
            loop                ; - - loop
              JSR  SRGetChar_   ; - - - SRGetChar()
              TSTB              ; - - - quit if key = NULLCHAR
              quif eq           ; - - -
              BSR  LocalEcho    ; - - - echo received char on screen
              TST  ReceiveLock  ; - - -
            until eq            ; - - until ~ ReceiveLock
          endif                 ; - endif
        endloop                 ; endloop
        LDB     #CTRL_S         ; transmit XOFF in case host has more for us
        JSR     SRPutChar_      ; SRPutChar( CTRL_S )
                                ; get out those last few characters
        loop                    ; loop
          JSR   SRGetChar_      ; - SRGetChar()
          TSTB                  ; - quit if character = NULLCHAR
          quif  eq              ; -
          BSR   LocalEcho       ; - echo received char on screen
        endloop                 ; endloop
        JSR     SRDisconnect_   ; SRDisconnect()
        LDD     KBIn            ; get keyboard file pointer
        JSR     Closef_         ; Closef( KBIn )
        CLR     Service         ; set exit code
        RTS                     ; end of terminal emulation

TranOut equ  *
        guess                   ; guess
          CMPB  #KBESC          ; - quit if Key <> KBESC
          quif  ne              ; -
          LDB   #ESCAPE         ; - Key := ESCAPE
        admit                   ; admit
          CMPB  #KBDEL          ; - quit if Key <> KBDEL
          quif  ne              ; -
          LDB   #RUBOUT         ; - Key := RUBOUT
        admit                   ; admit
          LDA   PrevKey         ; - quit if PrevKey <> KBCTRL
          CMPA  #KBCTRL         ; -
          quif  ne              ; -
          ANDB  #$1F            ; - Key := Key & $1F
        endguess                ; endguess
        RTS                     ; end of keyboard translation for output

LocalEcho  equ  *
        PSHS    B               ; save character
        CLRA                    ; zero hi part of D
        guess                   ; guess
          CMPB  #CR             ; - quit if Key <> CR
          quif  ne              ; -
          JSR   PutChar_        ; - PutChar( CR )
          LDB   #CURUP          ; - PutChar( CURUP )
          JSR   PutChar_        ; - counteract Linefeed
        admit                   ; admit
          CMPB  #CTRL_G         ; - quit if Key <> CTRL_G
          quif  ne              ; -
          BSR   Beep_           ; - ring the bell
        admit                   ; admit
          CMPB  #CTRL_N         ; - quit if Key <> CTRL_N
          quif  ne              ; -
          LDB   #APL            ; -
          JSR   TSetChar_       ; - APL character set selected
        admit                   ; admit
          CMPB  #CTRL_O         ; - quit if Key <> CTRL_O
          quif  ne              ; -
          LDB   #ASCII          ; -
          JSR   TSetChar_       ; - ASCII character set selected
        admit                   ; admit
          guess                 ; - guess APL character set & '_'
            TST SKD_+APLSET     ; - - quit if not APL character set
            quif eq             ; - -
            CMPB #'F'           ; - - quit if not APL '_'
            quif ne             ; - -
            LDX SKD_+CURSOR     ; - - get cursor offset
            LEAX Screen_,X      ; - - point at current character
            LDA  ,X             ; - - get character already there
            CMPA #'a'+$80       ; - - quit if character < APL A
            quif lo             ; - -
            CMPA #'z'+$80       ; - - quit if character > APL Z
            quif hi             ; - -
            TFR  A,B            ; - - use reverse video character
          endguess              ; - endguess
          CLRA                  ; - zero hi part of D
          JSR   PutChar_        ; - echo character on screen
        endguess                ; endguess
        PULS    B               ; restore Key
        RTS                     ; end of local echo routine

;
; BEEP the SuperPET Buzzer
;
; Registers :   A, B (D) destroyed
;               X        preserved
;
Freq    equ     $E848   ; frequency
Wave    equ     $E84A   ; waveform
Music   equ     $E84B   ; music on/off
;
xdef    Beep_
Beep_   equ  *
        PSHS    X       ; save x reg
        LDB     #16     ; on = 16
        STB     Music   ; turn on the music
        LDB     #$0F    ; get waveform 00001111 off/off/off/off/on/on/on/on
        STB     Wave    ; and set it
        LDX     #3      ; set counter
        loop            ; loop
          LDB   #55     ; - get frequency
          BSR   Play    ; - and play it
          LDB   #60     ; - get frequency
          BSR   Play    ; - and play it
          LEAX  -1,X    ; - bump counter
        until eq        ; until counter exhausted
        LDB     #00     ; off=00
        STB     Music   ; turn off the music
        PULS    X       ; restore X reg
        RTS             ; return to caller

Play    equ     *
        STB     Freq    ; set frequency
        LDD     #0      ; set counter
        loop            ; loop
          ADDD  #1      ; - increment
          CMPD  #$0BFF  ; - check for duration
        until eq        ; until duration over
        RTS             ; end of play routine

KeyBoard fcc     "KEYBOARD"
         fcb     0

ReadMode fcc    "R"
         fcb    0

KBIn    rmb     2       ; keyboard file pointer
Hold    rmb     1       ; hold screen (scroll/no scroll toggle)
PrevKey rmb     1       ; previously typed key

        end
